libp2p_uds

package
v0.18.7 Latest Latest
Warning

This package is not in the latest version of its module.

Go to latest
Published: Apr 7, 2026 License: MIT Imports: 14 Imported by: 0

Documentation

Overview

Package uds was auto-translated from rust-libp2p. https://github.com/libp2p/rust-libp2p/blob/master/transports/uds/src/lib.rs

Index

Constants

This section is empty.

Variables

This section is empty.

Functions

This section is empty.

Types

type UdsTransport

type UdsTransport struct {
	// contains filtered or unexported fields
}

UdsTransport is the Unix Domain Socket transport.

func NewUDSTransport

func NewUDSTransport(
	upgrader transport.Upgrader, rcmgr network.ResourceManager,
	opts ...func(*UdsTransport) error,
) (*UdsTransport, error)

NewUDSTransport creates a UDS transport object.

func (*UdsTransport) CanDial

func (t *UdsTransport) CanDial(addr ma.Multiaddr) bool

CanDial returns true if this transport believes it can dial the given multiaddr.

func (*UdsTransport) Dial

func (t *UdsTransport) Dial(
	ctx context.Context, raddr ma.Multiaddr, p peer.ID,
) (transport.CapableConn, error)

Dial dials the peer at the remote address.

func (*UdsTransport) DialWithUpdates

func (t *UdsTransport) DialWithUpdates(
	ctx context.Context, raddr ma.Multiaddr, p peer.ID,
	updateChan chan<- transport.DialUpdate,
) (transport.CapableConn, error)

func (*UdsTransport) Listen

func (t *UdsTransport) Listen(laddr ma.Multiaddr) (transport.Listener, error)

Listen listens on the given multiaddr.

func (*UdsTransport) Protocols

func (t *UdsTransport) Protocols() []int

Protocols returns the list of terminal protocols this transport can dial.

func (*UdsTransport) Proxy

func (t *UdsTransport) Proxy() bool

Proxy always returns false for the UDS transport.

func (*UdsTransport) String

func (t *UdsTransport) String() string

Jump to